module Logging where

import Prelude hiding (catch, read)

import KoL.Util
import KoL.UtilTypes

import System.IO
import Control.Applicative

import Data.Time.Clock

import Data.IORef

import Control.Exception

import Text.Printf
import Control.Concurrent
import Control.Monad

import qualified Data.Map

-- import System.IO.Unsafe

-- 	unsafeGlobalVar :: IORef Handle
-- 	{-# NOINLINE unsafeGlobalVar #-}
-- 	unsafeGlobalVar = unsafePerformIO $ do
-- 		h <- openFile "chan-debug.txt" AppendMode
-- 		newIORef h

-- {-# NOINLINE unsafeGlobalVarCounter #-}
-- unsafeGlobalVarCounter = unsafePerformIO $ do
-- 	newIORef 0.0

-- time_action msg action = do
-- 	start <- getCurrentTime
-- 	let diffms before now = fromRational $ 1000 * toRational (diffUTCTime now before) :: Double
-- 	padding <- atomicModifyIORef unsafeGlobalVar (\x -> (x ++ " ", x))
-- 	putStrLn $ printf "%s > {%s} %s" padding (show start) msg
-- 	x <- action
-- 	atomicModifyIORef unsafeGlobalVar (\x -> (tail x, ()))
-- 	end <- getCurrentTime
-- 	ctr <- atomicModifyIORef unsafeGlobalVarCounter (\x -> (x + (diffms start end), (x + (diffms start end))))
-- 	putStrLn $ printf "%s < {%s}[%8.2fms] %s" padding (show end) (diffms start end) msg
-- 	putStrLn $ printf "%s ctr %8.2fms" padding ctr
-- 	return x

-- time_action msg action = action

-- log_write_chan msg = ((do
-- 	t <- getCurrentTime
-- 	h <- readIORef unsafeGlobalVar
-- 	hPutStrLn h (printf "[%-35s] %s" (show t) (msg :: String))
-- 	hFlush h) `catch` (\e -> do
-- 		putStrLn $ "log_debug_msg error:" ++ (show (e :: Control.Exception.SomeException))
-- 		return ()))

holdit ref action = writeChan (logchan ref) action

print_log_msg ref file logdetails = do
	log_debug_msg ref $ "writing log: " ++ file
	(doDbLogAction $ sessionData $ ref) ref $ \db -> do
-- 		putStrLn $ "writing to log db."
		let showstate s = case s of
			Just (_requestmap, _sessionmap, charmap, ascmap, daymap) -> show [("character", charmap), ("ascension", ascmap), ("day", daymap)]
			_ -> throw $ InternalError $ "Invalid state while logging"
		do_db_query_ db "INSERT INTO pageloads(time, statusbefore, statusafter, statebefore, stateafter, sessionid, requestedurl, parameters, retrievedurl, pagetext) VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?, ?);" [
			Just $ show $ time $ logdetails,
			Just $ apiStatusBefore $ logdetails,
			Just $ apiStatusAfter $ logdetails,
			Just $ showstate (stateBefore $ logdetails),
			Just $ showstate (stateAfter $ logdetails),
			Just $ sessionId $ logdetails,
			Just $ show $ requestedUri $ logdetails,
			show <$> (parameters $ logdetails),
			Just $ show $ retrievedUri $ logdetails,
			Just $ pageText $ logdetails]

log_nothing msg = do
	return (msg :: String)
	return () :: IO ()

log_debug_msg _ref = log_nothing

-- log_debug_msg ref msg = do
-- 	t <- getCurrentTime
-- 	holdit ref $ ((do
-- 		h <- openFile "debug-log.txt" AppendMode
-- 		hPutStrLn h (printf "[%-35s] %s" (show t) (msg :: String))
-- 		hClose h) `catch` (\e -> do
-- 			putStrLn $ "log_debug_msg error:" ++ (show (e :: Control.Exception.SomeException))
-- 			return ()))

log_to_file ref filename msg = holdit ref $ do
	h <- openFile filename AppendMode
	hPutStrLn h msg
	hClose h

log_file_retrieval ref url params = do
	log_debug_msg ref $ "retrieve file: " ++ (show url)
	holdit ref $ do
		h <- openFile "files-downloaded.txt" AppendMode
		hPutStrLn h ((show url) ++ "  " ++ (show params))
		hClose h

internal_log_time_msg ref debugmsg msg = do
	log_debug_msg ref $ debugmsg
	holdit ref $ do
		h <- openFile "timing-log.txt" AppendMode
		hPutStrLn h $ msg
		hClose h

lua_log_line ref msg action = do
	start <- getCurrentTime
	let diffms before now = fromRational $ 1000 * toRational (diffUTCTime now before) :: Double
-- 	x <- log_time_interval ref ("lua:" ++ msg) $ action
	x <- action
	end <- getCurrentTime
	let str = printf "[%8.1fms] %s" (diffms start end) msg
	log_debug_msg ref $ "lua-log: " ++ str
	holdit ref $ ((do
		h <- openFile "lua-log.txt" AppendMode
		hPutStrLn h $ str
		hClose h) `catch` (\e -> do
			putStrLn $ "lua_log_line error:" ++ (show (e :: SomeException))
			return ()))
	return x 

diffms before now = fromRational $ 1000 * toRational (diffUTCTime now before) :: Double

log_retrieval ref msg start end = do
	ct <- getCurrentTime
	let str = printf "[%-35s] [%8.1fms] %s" (show ct) (diffms start end) msg
	writeChan (solid_logchan_ $ logstuff_ $ ref) $ do
		h <- openFile "http-log.txt" AppendMode
		hPutStrLn h $ str
		hClose h

log_time_uri ref msg uri = internal_log_time_msg ref ("logtime url: " ++ (show uri)) (printf "%s: %s" msg (show uri))

log_time_interval_above ref _threshold name action = do
	time_pre <- getCurrentTime
	indents <- readIORef (logindents ref)
	let indent_text = replicate (fromIntegral indents) ':'
	internal_log_time_msg ref ("logtime interval start: " ++ name) (printf "  [%-35s %s>] %s" (show time_pre) indent_text name)
	atomicModifyIORef (logindents ref) (\x -> (x + 1, ()))
	x <- (action) `catch` (\e -> do
		putStrLn $ "log:" ++ name ++ " exception: " ++ (show (e :: SomeException))
		throwIO e)
	atomicModifyIORef (logindents ref) (\x -> (x - 1, ()))
	time_post <- getCurrentTime
	let ms = diffms time_pre time_post
-- 	when (ms > threshold) $ do
	internal_log_time_msg ref ("logtime interval end: " ++ (printf "%s [%7.1fms]" name ms)) (printf "  [%-35s %s<] %s (%7.1fms)" (show time_post) indent_text name ms)
	return x

log_time_interval_http_above _ref _threshold name action = do
	x <- (action) `catch` (\e -> do
		putStrLn $ "log_http:" ++ name ++ " exception: " ++ (show (e :: SomeException))
		throwIO e)
	return x

log_time_interval ref name action = log_time_interval_above ref 10 name action

log_time_interval_http ref name action = log_time_interval_http_above ref 10 name action

log_page_result ref status_before log_time state_before uri params effuri pagetext status_after state_after = do
	(charname, charasc) <- do
		let getSessState var = do
			when (not $ stateValid_ $ ref) $ putStrLn $ "State invalid while logging and trying to get session state"
			Just (_, st) <- readIORef $ state ref
			let value = case st of
				(_requestmap, sessionmap, _charmap, _ascmap, _daymap) -> Data.Map.lookup var sessionmap
			return (value :: Maybe String)
		Just charname <- getSessState "character"
		Just charasc <- getSessState "ascension number"
		return (charname, charasc)
	let Just sessid = get_sessid $ cookie $ connection $ ref
	print_log_msg ref (charname ++ "-" ++ charasc ++ "-detailed.txt") $ LogItem { time = log_time, apiStatusBefore = status_before, apiStatusAfter = status_after, stateBefore = state_before, stateAfter = state_after, sessionId = sessid, requestedUri = uri, parameters = params, retrievedUri = effuri, pageText = pagetext }
